class Settings:
    """ A static class containing settings for all Config children. """
    
    frozen: bool = False
    """ Represents if a config should be frozen or not. """
    
    def __new__(cls, *args, **kwargs):
        raise ValueError("Cannot instantiate the Settings")

__all__ = \
[
    'Settings',
]



from enum import Enum
from typing import *

from dataclasses_json import LetterCase
from pyhocon import ConfigTree

_FIELD_CONSTRUCTOR_ARGUMENT_TYPE = '__config_constructor_argument_type__field_name__'
_FIELD_CONSTRUCTOR_POST_INIT = '__config_constructor_post_init__field_name__'

class ConstructorDataType(Enum):
    """
    A data type which is used for the constructor
    
    The following data types are allowed:
    
     - String
     - Integer
     - Float
     - Boolean
     - Config
     - List
    """
    
    String = ConfigTree.get_string
    Integer = ConfigTree.get_int
    Float = ConfigTree.get_float
    Boolean = ConfigTree.get_bool
    Config = ConfigTree.get_config
    List = ConfigTree.get_list
    Any = ConfigTree.get

T = TypeVar('T')
PostInitFunctionType = Callable[[T, Type[T]], Any]
"""
A type alias for
`(self, cls) => Optional[self]`
"""

def deserialize_with(constructor_data_type: ConstructorDataType, *, post_init: Union[PostInitFunctionType, str, None] = None) -> Callable[[Type[T]], Type[T]]:
    """
    Decorates a class to mark it as an automatically deserialized while reading the configuration.
    
    Args:
        constructor_data_type: `ConstructorDataType`. A data type which is used for the constructor.
        post_init: Optional. `str` or `dataclasses_config.decorations.PostInitFunctionType`.
            If presented, defines the post-init callback for that type.
            It would be called after the constructor, with the following arguments:
            
             - instance
             - class (exactly as described in type annotations)
             - a `dataclasses_config.config.Config` instance which calls the post-init.
            
            If this method returns something that is not `None`,
            the original value would be reassigned.
            
            If the argument value is the string rather than callable,
            then it is counted as the class's method name.
            That method signature should be the same.
    
    """
    
    def decorator(cls: Type['T']):
        setattr(cls, _FIELD_CONSTRUCTOR_ARGUMENT_TYPE, constructor_data_type)
        nonlocal post_init
        if (isinstance(post_init, str)):
            post_init = getattr(cls, post_init)
        setattr(cls, _FIELD_CONSTRUCTOR_POST_INIT, post_init)
        return cls
    return decorator
del T

def dataclass_json_config(*, letter_case: LetterCase = None, **kwargs):
    """
    Sets the config for the given JSON DataClass.
    
    Args:
        letter_case: `dataclasses_json.LetterCase`.
            A policy for JSON DataClass key switch.
        
        **kwargs:
            Other settings those are introduced in the future.
    """
    
    from dataclasses_json import config, DataClassJsonMixin
    
    if (letter_case is not None):
        kwargs['letter_case'] = letter_case
    
    D = TypeVar('T', bound=DataClassJsonMixin)
    def wrapper(cls: Type[D]) -> Type[D]:
        if (kwargs):
            cls.dataclass_json_config = config(**kwargs)['dataclasses_json']
        return cls
    del D
    
    return wrapper

__all__ = \
[
    'dataclass_json_config',
    'deserialize_with',
    
    'ConstructorDataType',
    'PostInitFunctionType',
]

__pdoc_extras__ = \
[
    'PostInitFunctionType',
]



from dataclasses import fields, Field, dataclass
from functools import partial, wraps
from typing import *

from dataclasses_json import Undefined, DataClassJsonMixin
from pyhocon import ConfigTree, HOCONConverter

def get_field(class_or_instance, field_name: str) -> Field:
    """
    Extracts the `dataclasses.Field` object from the class or instance metadata.
    
    Args:
        class_or_instance: a DataClass class or instance (`dataclasses.is_dataclass()` should return `True`)
        field_name: `str`. A field name to find.
    
    Returns:
        `dataclasses.Field`
    
    """
    
    try:
        _cls_f: Field = next(f for f in fields(class_or_instance) if f.name == field_name)
    except StopIteration as e:
        raise AttributeError(f"Dataclass {class_or_instance!r} does not have the attribute {field_name!r}.") from e
    else:
        return _cls_f

def dump_config(conf: ConfigTree, format: str = 'json', prefix: str = "Config dump:") -> str:
    """
    Dumps the specified `pyhocon.ConfigTree` in specified format.
    
    Args:
        conf: `pyhocon.ConfigTree` to dump.
        format: `str`.
            For the *pyhocon* release 0.4.0, the following formats are allowed:
            
             - `json` *(default)*
             - `properties`
             - `hocon`
             - `yaml`
        
        prefix: `str`. A string which would be printed before the config.
    
    Returns:
        `str`. A string representation of config in specified format.
    
    """
    lines = \
    [
        prefix,
        # '[START]',
        HOCONConverter.convert(conf, format),
        # '[END]',
    ]
    
    return '\n'.join(lines)

CLASS_WRAPS_UPDATES = \
[
    '__annotations__',
]
_wraps_class = wraps(wraps)(partial(wraps, updated=CLASS_WRAPS_UPDATES))
_wraps_class.__doc__ += \
"""
    This function is compatible with the classes, abstract classes and dataclasses wrapping
    in the newer Python versions.
    This function does not update `__dict__` by default, but updates `__annotations__` instead,
    which is required for the popper dataclasses work.
"""

@wraps(_wraps_class)
def wraps_class(*args, **kwargs):
    return _wraps_class(*args, **kwargs)

T = TypeVar('T')
@dataclass(frozen=True)
class DataclassesJsonMeta(DataClassJsonMixin):
    """
    A helper class representing a content of `dataclasses_json.config()` metadata structure.
    
    See Also:
        * `dataclasses.Field.metadata`
        * `dataclasses_json.cfg.config()`
    """
    
    encoder: Optional[Callable] = None
    decoder: Optional[Callable] = None
    mm_field: Any = None
    letter_case: Optional[Callable[[str], str]] = None
    undefined: Optional[Union[str, Undefined]] = None
    exclude: Optional[Callable[[str, T], bool]] = None


__all__ = \
[
    'dump_config',
    'get_field',
    'wraps_class',
    
    'CLASS_WRAPS_UPDATES',
    
    'DataclassesJsonMeta',
]



import warnings
from dataclasses import field
from importlib import import_module
from types import ModuleType
from typing import *

from .._helpers import extract_optional, extract_generic
from ..decorations import *

T = TypeVar('T')
@deserialize_with(ConstructorDataType.String, post_init='_check_class')
class DynamicClass(Generic[T]):
    """
    A helper class which accepts a string class path and imports it.
    Stores both class path and the loaded class.
    
    This class is the generic over `T` - super-class.
    If used inside `dataclasses_config.config.Config` class,
    it would also check that the loaded class is the subclass of `T`.
    
    Raises:
        ImportError: Raised in case of missing module or class in that module.
        TypeError: Raised in case of loaded class not matching the super-class.
    
    Warnings:
        ImportWarning: Sent if the inheritance check is called without specifying the base class.
    
    """
    
    class_path: str = field()
    """`str`. Full class path which could be used for import."""
    
    cls: Type[T]
    """`Type[T]`. An imported and loaded class."""
    
    def __init__(self, class_path: str, *, cls: Type[T] = None):
        self.class_path = class_path
        module_name, _sep, class_name = self.class_path.rpartition('.')
        if (not _sep):
            raise ImportError(f"Cannot import class '{self.class_path}'")
        
        if (cls is None):
            mod: ModuleType = import_module(module_name)
            try:
                self.cls = getattr(mod, class_name)
            except AttributeError as e:
                raise ImportError(f"Cannot import name {class_name!r} from {module_name!r} ({mod.__file__})") from e
        else:
            self.cls = cls
    
    @classmethod
    def from_cls(cls, target_cls: Type[T]) -> 'DynamicClass[T]':
        return cls(class_path=f'{target_cls.__module__}.{target_cls.__name__}', cls=target_cls)
    
    def _check_class(self, tp: Type['DynamicClass[T]'], *args, **kwargs):
        
        _, tp = extract_optional(tp)
        is_generic, base, type_args = extract_generic(tp)
        if (not is_generic):
            warnings.warn(ImportWarning(f"Type {tp} should be parametrized, but it is not."))
            return
        
        expected_parent = type_args[0]
        if (not issubclass(self.cls, expected_parent)):
            raise TypeError(f"{self.cls!r} is not a subclass of {expected_parent!r} (from {tp!r}).")
    
    def __repr__(self):
        return f'{type(self).__name__}({self.cls!r})'
del T

__all__ = \
[
    'DynamicClass',
]



import builtins
import os
from abc import ABC
from copy import copy
from dataclasses import fields, dataclass, Field, field, replace, MISSING
from enum import Enum
from functools import wraps
from logging import getLogger
from typing import *
from warnings import warn

import pkg_resources
from cacheable_iter import lru_iter_cache
from dataclasses_json import DataClassJsonMixin
from dataclasses_json.core import Json
from functional.filters import *
from functional.option import *
from pyhocon import ConfigTree, ConfigFactory, ConfigException, ConfigMissingException

from ._helpers import extract_generic, extract_optional
from .backports import cached_property
from .classes.path import Path, RelPath
from .decorations import *
# noinspection PyProtectedMember
from .decorations import _FIELD_CONSTRUCTOR_POST_INIT, _FIELD_CONSTRUCTOR_ARGUMENT_TYPE, ConstructorDataType
from .errors import *
from .settings import Settings
from .util import get_field, dump_config, wraps_class, DataclassesJsonMeta

logger = getLogger('dataclasses-config')
C = TypeVar('C', bound='Config')

_dataclass = wraps(dataclass)(dataclass)
HIDDEN_FIELD = '__hidden_field__'
DATACLASSES_JSON_META = 'dataclasses_json'

@overload
def _hidden_field(*, default: Any = MISSING, default_factory: Any = MISSING, hash: bool = None, metadata: dict = None) -> Field:
    pass
def _hidden_field(*, metadata: dict = None, **kwargs) -> Field:
    metadata = metadata or dict()
    metadata.setdefault(HIDDEN_FIELD, True)
    kwargs.setdefault('init', False)
    kwargs.setdefault('compare', False)
    kwargs.setdefault('repr', False)
    
    return field(metadata=metadata, **kwargs)

T = TypeVar('T')
@overload
def _cache(maxsize: int) -> Callable[[T], T]:
    pass
def _cache(*args, **kwargs):
    return lru_iter_cache(*args, **kwargs)

T = TypeVar('T')
@dataclass(frozen=True)
class FieldExtension(Generic[T]):
    """
    A helper class for dataclass field manipulations.
    Contains the original `dataclasses.Field` object,
    its positional index and type information.
    """
    
    base: Field
    """ An original `dataclasses.Field` object. """
    positional_index: int
    """ Positional index of the field in class (for getting/setting inside `*args`). """
    parent_cls: Type['Config'] = field(init=True, repr=False, compare=False, hash=False)
    """ The parent class containing this field. """
    
    is_optional: bool = field(init=False)
    """
    `True` only if the original type was a `None`-containing `Union`, including `Optional[T]`.
    Optional types are unwrapped for the purpose of `FieldExtension.tp`.
    """
    
    is_generic: bool = field(init=False)
    """
    `True` only if the original type was a subscripted `Generic`, i.e. `MyClass[int]`.
    Generic types are unwrapped for the purpose of `FieldExtension.tp`.
    """
    
    tp: Type[T] = field(init=False)
    """
    The unwrapped type.
    It has removed generics (i.e., `MyClass[int]` becomes `MyClass`),
    as well as removed optional (i.e., `Optional[MyClass]` becomes `MyClass`).
    """
    
    def __post_init__(self):
        tp = self.base.type
        is_optional, tp = extract_optional(tp)
        is_generic, tp, _ = extract_generic(tp)
        
        object.__setattr__(self, 'is_optional', is_optional)
        object.__setattr__(self, 'is_generic', is_generic)
        object.__setattr__(self, 'tp', tp)
    
    @cached_property
    def dcj_meta(self) -> DataclassesJsonMeta:
        """ A structured representation of metadata of the DataClass field of the DataClassesJSON package. """
        
        meta = self.base.metadata
        
        if (DATACLASSES_JSON_META in meta):
            data = meta[DATACLASSES_JSON_META]
        else:
            data = dict()
        
        if (self.parent_cls.dataclass_json_config is None):
            root = dict()
        else:
            root = copy(self.parent_cls.dataclass_json_config)
        
        root.update(data)
        return DataclassesJsonMeta.from_dict(root)
    
    @cached_property
    def kvs_name(self) -> str:
        """
        A name of field that should be used in the loader.
        
        Usually, it is the same as the `FieldExtension.base` `.name`,
        but overriding `dataclasses_json` field metadata
        
        See Also:
            * `FieldExtension.dcj_meta`
            * `dataclasses_json.cfg.config()`
        """
        
        if (self.dcj_meta.letter_case is not None):
            return self.dcj_meta.letter_case(self.base.name)
        else:
            return self.base.name
    
    @classmethod
    def from_class(cls, class_or_instance, filter: Union[Callable[['FieldExtension[T]'], bool], Type[T], None] = None) -> Iterator['FieldExtension[T]']:
        """
        Extracts the `dataclasses_config.config.FieldExtension`s from the given class.
        Returns an iterator.
        Ignores non-init fields.
        
        Args:
            class_or_instance: A DataClass type or DataClass instance.
            
            filter: Any of the following:
                
                - `None` *(default)*
                Nothing is filtered.
                
                - `Callable` with the signature:
                `dataclasses_config.config.FieldExtension` => `bool`
                This case, this function applies as a filter.
        
        """
        
        if (isinstance(class_or_instance, type)):
            target_cls = class_or_instance
        else:
            target_cls = type(class_or_instance)
        return cls._from_class(target_cls, filter)
    
    @classmethod
    @_cache(maxsize=1024)
    def _from_class(cls, target_cls, filter: Union[Callable[['FieldExtension[T]'], bool], Type[T], None]):
        
        fields_seq = builtins.filter(lambda f: f.init, fields(target_cls))
        seq = (cls(base=f, positional_index=i, parent_cls=target_cls) for i, f in enumerate(fields_seq))
        
        if (filter is None):
            return seq
        
        elif (isinstance(filter, type)):
            _base = filter
            filter = lambda f: isinstance(f.tp, type) and issubclass(f.tp, _base)
        elif (callable(filter)):
            pass
        else:
            raise TypeError(f"Invalid type of filter argument: {filter!r} ({type(filter)!r}). Expected: type, callable or None")
        
        return builtins.filter(filter, seq)

@dataclass(frozen=True)
class FieldExtensionTypeFilter(AbstractFilter[FieldExtension[T]], Generic[T]):
    """
    Applies the given filter to the `FieldExtension.tp` property and returns the result.
    """
    
    type_filter: AbstractFilter[Type[T]]
    """ A filter that should be applied to the `FiledExtension.tp`. """
    
    def check_element(self, el: FieldExtension[T]) -> bool:
        return self.type_filter(el.tp)

del T

@overload
def config_class \
(
    cls: Type[C] = None,
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    unsafe_hash: bool = False,
    frozen: bool = Settings.frozen,
    match_args: bool = True,
    kw_only: bool = False,
    slots: bool = False
) -> Type[C]:
    pass

def config_class(cls: Type[C] = None, **kwargs) -> Type[C]:
    kwargs.setdefault('frozen', Settings.frozen)
    # noinspection PyArgumentList
    return dataclass(cls, **kwargs)

T = TypeVar('T')
CONFIG_TREE_READERS: List[Tuple[Type[T], ConstructorDataType]] = \
[
    (dict, ConstructorDataType.Config),
    (list, ConstructorDataType.List),
    (bool, ConstructorDataType.Boolean),
    (int,  ConstructorDataType.Integer),
    (float,ConstructorDataType.Float),
    (str,  ConstructorDataType.String),
]
"""
A global mapping storing registered basic readers for ConfigTree for basic types.
Each is a function `get_T(self: ConfigTree, key: str) -> T`.
"""
del T

@config_class
# noinspection PyDataclass
class Config(DataClassJsonMixin, ABC):
    """
    A base class for the configuration.
    Could be automatically parsed from dict, json string and `pyhocon.ConfigTree`.
    
    Automatically processes the nested classes.
    Automatically post-processes classes marked with the `dataclasses_config.decorations.deserialize_with()` decorator.
    """
    
    _config: ConfigTree = _hidden_field(default=None)
    _config_kwargs: Dict[str, Any] = _hidden_field(default_factory=dict)
    
    @classmethod
    def from_config(cls: Type[C], config: ConfigTree, **kwargs) -> C:
        """
        Class-method.
        Parses the given `pyhocon.ConfigTree` and returns the new instance of this class.
        """
        
        res = cls.from_dict(config, config_kwargs=kwargs)
        object.__setattr__(res, '_config', config)
        object.__setattr__(res, '_config_kwargs', kwargs)
        return res
    
    @classmethod
    def from_dict(cls: Type[C], kvs: Json, *, config_kwargs: Dict[str, Any] = None, **kwargs) -> C:
        """
        Class-method.
        Parses the given parsed JSON-object and returns the new instance of this class.
        
        Args:
            kvs: `Dict[str, Any]`
            config_kwargs: Optional. `Dict[str, Any]`. Internal.
                `config_kwargs` are passed directly to the `dataclasses_config.config.Config.from_config()` calls
                while parsing the nested `Config`s.
            **kwargs: Extra options from the original `dataclasses_json.DataClassJsonMixin.from_dict()`.
        
        Raises:
            ConfigMissingKeyError: Raised when the config key is required but missing.
        
        """
        
        if (config_kwargs is None):
            config_kwargs = dict()
        
        config_tree: Optional[ConfigTree]
        if (isinstance(kvs, ConfigTree)):
            config_tree = kvs
        else:
            config_tree = ConfigTree(kvs)
        
        kvs = dict()
        for f in FieldExtension.from_class(cls):
            tp = f.tp
            kvs_name = f.kvs_name
            if (isinstance(tp, type)):
                cdt: ConstructorDataType = getattr(f.tp, _FIELD_CONSTRUCTOR_ARGUMENT_TYPE, None)
                if (cdt is not None):
                    for t, _cdt in CONFIG_TREE_READERS:
                        if (cdt is _cdt):
                            tp = t
                            break
                
                if (issubclass(tp, Config)):
                    tree = config_tree.get_config(kvs_name, MISSING)
                    if (tree is not MISSING):
                        kvs[kvs_name] = tp.from_config(tree, **config_kwargs)
                    continue
                
                elif (issubclass(tp, Enum)):
                    value = config_tree.get(kvs_name, MISSING)
                    if (value is MISSING):
                        continue
                    
                    updated = None
                    if (isinstance(value, str)):
                        try:
                            # noinspection PyUnresolvedReferences
                            updated = tp[value]
                        except ValueError:
                            pass
                    
                    if (updated is None):
                        # noinspection PyArgumentList
                        updated = tp(value)
                    
                    kvs[kvs_name] = updated
                    continue
                
                else:
                    if (cdt is None):
                        for t, _cdt in CONFIG_TREE_READERS:
                            if (issubclass(tp, t)):
                                cdt = _cdt
                                break
                    
                    if (cdt is not None):
                        try:
                            cdt: Callable[[ConfigTree, str], T]
                            v = cdt(config_tree, kvs_name)
                        except ConfigException as e:
                            try:
                                actual = config_tree.get(kvs_name)
                            except ConfigMissingException:
                                if (f.base.default is not MISSING or f.base.default_factory is not MISSING):
                                    continue
                            else:
                                raise ConfigWrongPropertyTypeError(cls, config_tree, kvs_name, tp, actual) from e
                            raise
                        else:
                            kvs[kvs_name] = v
                            continue
            
            warn(UnsupportedTypeWarning(cls, config_tree, kvs_name, tp), 2)
            kvs[kvs_name] = config_tree.get(kvs_name)
        
        try:
            kvs['_config_kwargs']=config_kwargs
            return super().from_dict(kvs, **kwargs)
        except KeyError as e:
            # noinspection PyArgumentList, PyTypeChecker
            raise ConfigMissingKeyError(cls, kvs, e.args[0]) from e
    
    def __post_init__(self):
        for f in FieldExtension.from_class(self, FieldExtensionTypeFilter(HasAttrFilter(_FIELD_CONSTRUCTOR_ARGUMENT_TYPE))):
            v = getattr(self, f.base.name)
            
            # ToDo:
            #  - Pass self._config_kwargs through dataclasses constructor
            #  - Remove Enum workaround
            if (v is None and (f.is_optional or self._config_kwargs.get('infer_missing', False) or issubclass(f.tp, Enum))):
                continue
            
            # ToDo: Check ConstructorDataType
            cdt: ConstructorDataType = getattr(f.tp, _FIELD_CONSTRUCTOR_ARGUMENT_TYPE)
            post: PostInitFunctionType = getattr(f.tp, _FIELD_CONSTRUCTOR_POST_INIT)
            
            v = f.tp(v)
            if (post is not None):
                x = post(v, f.base.type)
                if (x is not None):
                    v = x
            
            object.__setattr__(self, f.base.name, v)
    
    # noinspection PyDataclass
    def replace(self, **changes):
        """
        Return a new object replacing specified fields with new values.
        
        This is especially useful for frozen classes.
        
        Note:
            Hidden fields (like `Config`'s `._config`) are kept as-is,
            without any modification, so try to avoid using them mindlessly.
        
        Example:
            ```python
            @dataclass(frozen=Settings.frozen)
            class C(Config):
                x: int
                y: int
            
            c = C(1, 2)
            c1 = c.replace(x=3)
            assert c1.x == 3 and c1.y == 2
            ```
        """
        
        result: Config = replace(self, **changes)
        # ToDo:
        #  (?) Create copies hidden fields, instead of copying the links
        #  (?) Update values of hidden fields according to **changes
        
        for f in fields(result): # type: Field
            if (f.metadata.get(HIDDEN_FIELD, False)):
                object.__setattr__(result, f.name, getattr(self, f.name))
        
        return result
    
    D = TypeVar('D', bound=dict)
    def asdict(self, *, dict_factory: Type[D] = dict) -> D:
        """
        Return the fields of a dataclass instance as a new dictionary mapping
        field names to field values.
        
        Args:
            dict_factory: `Type[dict]`.
                If given, `dict_factory` will be used instead of built-in dict.
                The function applies recursively to field values that are
                dataclass instances. This will also look into built-in containers:
                tuples, lists, and dicts.
                
                Default: `dict`
        
        Example:
            ```python
            @dataclass(frozen=Settings.frozen)
            class C(Config):
                x: int
                y: int
            
            c = C(1, 2)
            assert c.asdict() == { 'x': 1, 'y': 2 }
            ```
        """
        
        return dict_factory(self._asdict_iter())
    
    def _asdict_iter(self) -> Iterator[Tuple[str, Any]]:
        for f in FieldExtension.from_class(self):
            yield f.base.name, getattr(self, f.base.name)
    
    del D

def main_config \
(
    *args,
    resources_directory: str = 'resources/',
    default_config_name: str = 'application.conf',
    reference_config_name: str = 'reference.conf',
    env_variable_name: Optional[str] = 'PYTHON_APPLICATION_CONFIG_PATH',
    root_config_path: str = None,
    module: str = None,
    log_config: bool = True,
) -> Callable[[Type[C]], Type[C]]:
    """
    Decorates the given `dataclasses_config.config.MainConfig` class with the default overrides.
    Multiple `dataclasses_config.config.main_config` directives override each other,
    and only the last is used.
    
    Args:
        *args:
        resources_directory: `str`.
            Defines where to look for the config files.
            This applies for both current directory lookup and module resources directory lookup.
            
            *Default: `resources/`*
        
        default_config_name: `str`
            Defines the default application config filename.
            It is searched in current directory and prioritized over the reference config
            when loading configuration via `dataclasses_config.config.MainConfig.default()`.
            
            *Default: `application.conf`*
        
        reference_config_name: `str`
            Defines the default reference config filename.
            It is searched in both current directory and resources directory
            when loading configuration via `dataclasses_config.config.MainConfig.default()`.
            Reference config values are fallback for values missing in the application config.
            
            *Default: `reference.conf`*
        
        env_variable_name: `Optional[str]`
            Defines the environment variable name to be used for searching default configuration.
            Could be set to None if this config could not be loaded via the environment variable.
            
            *Default: `PYTHON_APPLICATION_CONFIG_PATH`*
             
        root_config_path: Optional. `str`
            Defines the offset from the config root when loading the default and/or reference values.
            Ignored when calling `dataclasses_config.config.Config.from_config`
            
            By default, the root is used.
        
        module: Optional. `str`
            The module name resources are loaded from.
            (Reference config is loaded from the resources directory -- see above.)
            
            By default, the module from this class is used (which is get by `cls.__module__`).
        
        log_config: `bool`.
            If `True` (default), the config is logged to the `'dataclasses-config'` log
            with the INFO logging level.
    
    """
    
    def decorator(cls: Type[C]) -> Type[C]:
        @wraps_class(cls)
        @dataclass(frozen=Settings.frozen)
        class wrapper(dataclass(frozen=Settings.frozen)(cls)):
            _resources_directory: str = _hidden_field(default=resources_directory)
            _default_config_name: str = _hidden_field(default=default_config_name)
            _reference_config_name: str = _hidden_field(default=reference_config_name)
            _env_variable_name: str = _hidden_field(default=env_variable_name)
            _root_config_path: str = _hidden_field(default=root_config_path)
            _module: str = _hidden_field(default=module)
            _log: bool = _hidden_field(default=log_config)
        
        return wrapper
    
    if (args):
        return decorator(*args)
    else:
        return decorator

@main_config
class MainConfig(Config, ABC):
    """
    A `dataclasses_config.config.Config` extension which should be used
    for an application and/or module root config.
    
    This class **always** tries to load the reference config,
    whether the default config is loaded or the specific one.
    
    This reference config is then looked in current directory
    and module resources directory as a fallback.
    When loaded, it is merged with the loaded application config,
    with the application config's values prioritized.
    
    This works very similarly to the Scala's [lightbend/config](https://github.com/lightbend/config)
    
    ## See Also:
    1. Lightbend Config: https://github.com/lightbend/config
    2. Its Standard Behaviour: https://github.com/lightbend/config#standard-behavior
    """
    
    @classmethod
    def _get_module(cls) -> str:
        module: Optional[str] = get_field(cls, '_module').default
        if (module is None):
            module: str = cls.__module__
        
        return module
    
    @classmethod
    def _get_default_config_name(cls) -> str:
        return get_field(cls, '_default_config_name').default
    
    @classmethod
    def _get_reference_config_name(cls) -> str:
        return get_field(cls, '_reference_config_name').default
    
    @classmethod
    def _get_resources_directory(cls) -> str:
        return get_field(cls, '_resources_directory').default
    
    @classmethod
    def _get_env_variable_name(cls) -> Optional[str]:
        return get_field(cls, '_env_variable_name').default
    
    @classmethod
    def _get_log(cls) -> bool:
        return get_field(cls, '_log').default
    
    @classmethod
    def _get_root_config_path(cls) -> Optional[str]:
        return get_field(cls, '_root_config_path').default
    
    @classmethod
    def default(cls: Type[C]) -> C:
        """
        Tries to load the default config and merge reference config into it.
        
        If the `PYTHON_APPLICATION_CONFIG_PATH` environment variable exists and non-empty,
        then this would try to load config from this variable.
        Path might be relative and absolute.
        Raises an exception if config is missing.
        
        If such variable does not exist (or has empty value),
        then the default application config is searched.
        Search includes only resources directory in current working directory.
        Does not raise an exception if nothing is found.
        
        See `dataclasses_config.config.main_config()` for more information.
        
        Returns:
            `MainConfig`
        """
        
        default_config_name: str = cls._get_default_config_name()
        resources_directory: str = cls._get_resources_directory()
        
        env_var_name = cls._get_env_variable_name()
        application_conf_path = None
        if (env_var_name is not None):
            application_conf_path = os.environ.get(env_var_name, None)
        if (application_conf_path):
            ignore_missing = False
        else:
            application_conf_path = os.path.join(resources_directory, default_config_name)
            ignore_missing = True
        
        return cls.load_config(application_conf_path, ignore_missing=ignore_missing)
    
    @classmethod
    @overload
    def load_config(cls: Type[C], path: str, *, ignore_missing: bool = False, use_reference_conf: bool = True) -> C:
        pass
    
    @classmethod
    def load_config(cls: Type[C], path: str, *, ignore_missing: bool = False, use_reference_conf: bool = MISSING) -> C:
        """
        Loads the given application config and merges the reference config into it.
        Applies root config offset if any.
        
        Args:
            path: `str`.
                A path to load config from.
                Path might be relative and absolute.
            
            ignore_missing: `bool`.
                Defines whether to raise an exception if config could not be loaded (`False`, default) or not (`True`).
            
            use_reference_conf: `bool`.
                 If set to `True` (default behaviour), reference config will be merged into the config before instantiating
        
        Returns:
            `MainConfig`
        """
        
        kwargs = dict()
        if (use_reference_conf is not MISSING):
            kwargs['use_reference_conf'] = use_reference_conf
        
        root_config_path: Optional[str] = cls._get_root_config_path()
        
        # application.conf
        # It is environment-dependent
        try:
            application_conf: ConfigTree = ConfigFactory.parse_file(path)
        except FileNotFoundError:
            if (ignore_missing):
                application_conf = ConfigTree()
            else:
                raise
        
        if (root_config_path is not None):
            application_conf = application_conf.get_config(root_config_path, ConfigTree())
        
        return cls.from_config(application_conf, **kwargs)
    
    @classmethod
    def reference_config(cls) -> ConfigTree:
        """
        Returns the `pyhocon.ConfigTree` representing the content of reference config.
        Reference config is searched in resources directories at
        current working directory (first) and module directory (next).
        Returns empty config if nothing found.
        Applies root config offset if any.
        
        Returns:
            `pyhocon.ConfigTree`
        """
        
        reference_config_name: str = cls._get_reference_config_name()
        module: str = cls._get_module()
        resources_directory: str = cls._get_resources_directory()
        root_config_path: Optional[str] = cls._get_root_config_path()
        
        # reference.conf
        # It is built inside package resources
        try:
            expected = os.path.join(resources_directory, reference_config_name)
            reference_conf_path: str = pkg_resources.resource_filename(module, expected)
            if (not os.path.isfile(reference_conf_path)):
                reference_conf_path = expected
            reference_conf: ConfigTree = ConfigFactory.parse_file(reference_conf_path)
        except FileNotFoundError:
            reference_conf = ConfigTree()
        else:
            del reference_conf_path, expected
        
        if (root_config_path is not None):
            reference_conf = reference_conf.get_config(root_config_path, ConfigTree())
        
        return reference_conf
    
    @classmethod
    def from_config(cls: Type[C], config: ConfigTree, *, use_reference_conf: bool = True, **kwargs) -> C:
        """
        Same as the super-class' `dataclasses_config.config.Config.from_config()`,
        but merges the reference config.
        
        Optionally logs the config if such option was provided in
        `dataclasses_config.config.main_config()` parameters.
        """
        
        log: bool = cls._get_log()
        
        if (use_reference_conf):
            reference_conf = cls.reference_config()
            # The resulting config is the result of merging of reference config and application config
            config = ConfigTree.merge_configs(reference_conf, config, copy_trees=True)
        
        if (log):
            logger.info(dump_config(config))
        return super().from_config(config, **kwargs)

@dataclass(frozen=Settings.frozen)
class ConfigWithRoot(Config, ABC):
    """
    Extension to the `dataclasses_config.config.Config` which
    allows `dataclasses_config.classes.path.RelPath` post-processing.
    """
    
    root: Optional[Path]
    """ Optional. `str`. Current root directory which is used for `dataclasses_config.classes.path.RelPath` instances. """
    
    def with_root(self: C, root: Optional[Path] = None, *, default_path: str = '.', **kwargs) -> C:
        """
        Creates a copy of this class, with all instances of
        `dataclasses_config.classes.path.RelPath` transformed to their absolute path form.
        (But still remain instances of `dataclasses_config.classes.path.RelPath`)
        
        Overrides `root` field in the resulting instance.
        Should work fine with the frozen dataclasses (potentially).
        
        The first available root is applied:
        1. `root` argument.
        2. `root` field.
        3. `default_path` argument.
        
        Args:
            root: `str`.
                A first-priority root to apply.
            
            default_path: `str`.
                A fallback root to apply.
                
                *Default: `.`*
            
            **kwargs:
                Format arguments for `dataclasses_config.classes.path.RelPath.apply()`
        
        Returns:
        
        """
        
        root: Path = (Option(root) or Option(self.root)).get_or_else(Path(default_path))
        
        kvs = dict()
        for f in FieldExtension.from_class(self, RelPath):
            v = getattr(self, f.base.name)
            if (isinstance(v, RelPath)):
                kvs[f.base.name] = v.apply(root, **kwargs)
        
        kvs.setdefault('root', root)
        return self.replace(**kvs)


__all__ = \
[
    'CONFIG_TREE_READERS',
    
    'config_class',
    'main_config',
    
    'Config',
    'ConfigInvalidError',
    'ConfigMissingKeyError',
    'ConfigWithRoot',
    'FieldExtension',
    'FieldExtensionTypeFilter',
    'MainConfig',
]
__pdoc_extras__ = \
[
    'CONFIG_TREE_READERS',
]



import sys

if (sys.version_info >= (3, 8)):
    from functools import cached_property
else:
    # noinspection PyUnresolvedReferences
    from cached_property import cached_property


__all__ = \
[
    'cached_property',
]



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import *

from pyhocon import ConfigException, ConfigTree, ConfigWrongTypeException

from dataclasses_config.util import dump_config

# noinspection PyUnreachableCode
if (False):
    # This module is always loaded after `dataclasses_config.config`
    # and thus this import statement cannot be called properly.
    # However, here it's required only for type annotations, so this is fine.
    from .config import Config

@dataclass(frozen=True)
class DataclassExceptionABC(Exception, ABC):
    """
    Base *abstract* exception class used inside this package.
    """
    
    @property
    @abstractmethod
    def message(self) -> str:
        """ An exception message. """
        pass
    
    def __post_init__(self):
        super(DataclassExceptionABC, self).__init__(self.message)
    
    def __hash__(self):
        return hash(self.message)


@dataclass(frozen=True)
class ConfigInvalidError(ConfigException, DataclassExceptionABC):
    """
    Base exception class used inside this package for Config classes.
    """
    
    cls: Type['Config']
    """ A `dataclasses_config.config.Config` which caused an exception. """
    
    config: ConfigTree
    """ A `pyhocon.ConfigTree` which caused an exception. """
    
    @property
    def dump(self) -> str:
        """ A config string representation. """
        return dump_config(self.config)
    
    @property
    def message(self) -> str:
        return f"Config for class {self.cls.__name__!r} is invalid. {self.dump}"

@dataclass(frozen=True)
class ConfigMissingKeyError(ConfigInvalidError):
    """
    This exception is raised when the config key is requested but missing.
    This exception **is not** a successor of `KeyError`
    due to the way `KeyError`s are represented.
    """
    
    key: str
    """ A config key which caused an exception. """
    
    @property
    def message(self) -> str:
        return f"Config for class {self.cls.__name__!r} does not contain a required field {self.key!r}. {self.dump}"

@dataclass(frozen=True)
class ConfigWrongPropertyTypeError(ConfigInvalidError, ConfigWrongTypeException, TypeError):
    """
    This exception is raised when the config key is requested with wrong associated type,
    i.e. when a string was discovered when integer was expected.
    This also respects `dataclasses_config.decorations.deserialize_with()` decoration.
    """
    
    key: str
    """ A config key which caused an exception. """
    
    expected_type: type
    """ The type that was expected to be extracted. """
    
    actual_value: Any
    """ Actual extracted value. Type is calculated dynamically. """
    
    @property
    def message(self) -> str:
        return f"Config for class {self.cls.__name__!r} has wrong value type for field {self.key!r}. Expected: {self.expected_type.__name__!r}, got: {type(self.actual_value).__name__!r} ({self.actual_value!r})"

@dataclass(frozen=True)
class UnsupportedTypeWarning(ConfigInvalidError, Warning, DataclassExceptionABC):
    """
    This warning is raised when Config class unsuccessfully discovered the field type
    and thus was forced to use generic, non-type-safe loader for value.
    """
    
    key: str
    """ A config key which caused an exception. """
    
    tp: Optional[type] = None
    """
    Optional.
    The type that was expected to be extracted.
    `None` means that no types were associated with that field.
    """
    
    @property
    def message(self) -> str:
        base = f"Config for class {self.cls.__name__!r} was unable to discover proper loader for field {self.key!r}."
        if (self.tp is None):
            return base + ' ' + f"No types were associated with that field."
        else:
            return base + ' ' + f"Type {self.tp.__name__!r} has no associated loader."


__all__ = \
[
    'DataclassExceptionABC',
    'ConfigWrongPropertyTypeError',
    'ConfigInvalidError',
    'ConfigMissingKeyError',
    'UnsupportedTypeWarning',
]



from typing import *

from typing_inspect import is_generic_type, get_origin, is_optional_type, get_args

T = TypeVar('T')
def extract_optional(tp: Union[Type[Optional[T]], Type[T]]) -> Tuple[bool, Type[T]]:
    """
    Helper function that checks if the given type is optional,
    and if it is, expands it.
    Correctly handles both `Union[..., None]` and `Optional[...]` cases.
    
    Args:
        tp: `Type[T]` - potentially optional type.
    
    Returns:
        Returns a tuple of 2 values.
        
        - `bool`: **True** if the given type is `Optional`; **False** otherwise.
        - `Type[R]`: The class *T*, if *T* is not optional; and *R* if *T* is `Optional[R]`.
    
    Examples:
        ```python
        extract_optional(dict)                      # => (False, dict)
        extract_optional(Optional[int])             # => (True,  int)
        extract_optional(MyClass[str])              # => (False, MyClass[str])
        extract_optional(Union[MyClass[str], None]) # => (True,  MyClass[str])
        ```
    """
    
    if (is_optional_type(tp)):
        tp = get_args(tp, evaluate=True)[0]
        return True, tp
    else:
        return False, tp

def extract_generic(tp: Type[T]) -> Tuple[bool, Type[T], Tuple[Type, ...]]:
    """
    Helper function that checks if the given type is generic,
    and if it is, expands it.
    
    Args:
        tp: `Type[T]` - potentially generic type.
    
    Returns:
        Returns a tuple of 3 values.
        
        - `bool`: **True** if the given type is `Generic`; **False** otherwise.
        - `Type[R]`: The class *T*, if *T* is not optional; and *R* if *T* is `Optional[R]`.
        - `Tuple[Type, ...]`: The tuple of class parameters used for *T*'s creation; empty tuple if it is not generic.
    
    Examples:
        ```python
        extract_generic(dict)                      # => (False, dict,    ())
        extract_generic(Dict[A, B])                # => (True,  Dict,    (A, B))
        extract_generic(Optional[int])             # => (True,  Union,   (int, None))
        extract_generic(MyClass[str])              # => (True,  MyClass, (str))
        extract_generic(Union[MyClass[str], None]) # => (True,  Union,   (MyClass[str], None))
        ```
    """
    
    if (is_generic_type(tp)):
        base = get_origin(tp)
        return True, base, get_args(tp, evaluate=True)
    else:
        return False, tp, tuple()
del T

def create_documentation_index(submodules: list, __name__: str, __pdoc__: Dict[str, Union[str, bool]], __pdoc_extras__: List[str]):
    for submodule in submodules:
        submodule_name = submodule.__name__.partition(f'{__name__}.')[-1]
        __pdoc__[submodule_name] = True
        submodule_extras = getattr(submodule, '__pdoc_extras__', list())
        for _element in submodule.__all__:
            __pdoc__[_element] = _element in submodule_extras
        __pdoc_extras__.extend(submodule_extras)


__all__ = \
[
    'create_documentation_index',
    'extract_generic',
    'extract_optional',
]



"""
.. include:: ../../README.md
"""

from collections import namedtuple

__title__ = 'dataclasses-config'
__author__ = 'Peter Zaitcev / USSX Hares'
__license__ = 'BSD 2-clause'
__copyright__ = 'Copyright 2019-2022 Peter Zaitcev'
__version__ = '0.4.8'

VersionInfo = namedtuple('VersionInfo', 'major minor micro releaselevel serial')
version_info = VersionInfo(*__version__.split('.'), releaselevel='alpha', serial=0)

from .classes import *
from .config import *
from .decorations import *
from .settings import *

__all__ =  \
[
    'version_info',
    '__title__',
    '__author__',
    '__license__',
    '__copyright__',
    '__version__',
]
__pdoc__ = { }
__pdoc_extras__ = [ ]

submodules = \
[
    classes,
    config,
    decorations,
    settings,
]

for _m in submodules: __all__.extend(_m.__all__)
from ._helpers import create_documentation_index
create_documentation_index(submodules, __name__, __pdoc__, __pdoc_extras__)
del create_documentation_index, submodules



import os

from ..decorations import *

@deserialize_with(ConstructorDataType.String)
class Path(str):
    r"""
    A helper class which converts any string-like path to the absolute path.
    
    It is a subclass of `str`, and thus any string operations and methods are available here.
    Useful while reading configuration.
    Automatically converted to the absolute path when used inside `dataclasses_config.config.Config`.
    
    Examples:
        ```python
        Path('.')
        # Something like:
        #   '/home/user/project/test'
        #   'C:\\Users\\user\\Projects\\Test'
        
        Path('my_lib', 'mod.py')
        # Something like:
        #   '/home/user/project/test/my_lib/mod.py'
        #   'C:\\Users\\user\\Projects\\Test\\my_lib\\mod.py'
        
        ```
    """
    
    def __new__(cls, *args, **kwargs) -> 'Path':
        # noinspection PyTypeChecker
        return os.path.abspath(os.path.join(*args))

@deserialize_with(ConstructorDataType.String)
class RelPath(str):
    """
    A helper class which represents relative path.
    A method `dataclasses_config.classes.path.RelPath.apply()` converts this this object to the `dataclasses_config.classes.path.Path` with the given root.
    
    It is a subclass of `str`, and thus any string operations and methods are available here.
    Useful while reading configuration.
    Automatically converted to the absolute path when used inside `dataclasses_config.config.Config`.
    """
    
    def apply(self, root: Path, **kwargs) -> Path:
        r"""
        Converts this this object to the `dataclasses_config.classes.path.Path` with the given root.
        
        Args:
            root: `dataclasses_config.classes.path.Path`.
            **kwargs: Optional.
                If presented, then the resulting Path is then formatted using the `str.format_map()` method.
        
        Returns:
            `dataclasses_config.classes.path.Path`
        
        Examples:
            ```python
            RelPath('mod.py').apply('.')
            # Something like:
            #   '/home/user/project/test/mod.py'
            #   'C:\\Users\\user\\Projects\\Test\\mod.py'
            
            RelPath('{module_name}/mod.py').apply('my_lib')
            # Something like:
            #   '/home/user/project/test/my_lib/{module_name}/mod.py'
            #   'C:\\Users\\user\\Projects\\Test\\my_lib\\{module_name}\\mod.py'
            
            RelPath('{module_name}/mod.py').apply('my_lib', module_name=bin)
            # Something like:
            #   '/home/user/project/test/my_lib/bin/mod.py'
            #   'C:\\Users\\user\\Projects\\Test\\my_lib\\bin\\mod.py'
            ```
        """
        
        p = Path(root, self)
        if (kwargs):
            p = Path(p.format_map(kwargs))
        return p
    
    @property
    def as_path(self) -> Path:
        """
        Transforms the `RelPath` object to the absolute `dataclasses_config.classes.path.Path`.
        Returns:
            `dataclasses_config.classes.path.Path`
        """
        
        return Path(self)

__all__ = \
[
    'Path',
    'RelPath',
]



# noinspection PyProtectedMember
from setuptools import _install_setup_requires
_install_setup_requires(dict(setup_requires=[ 'extended-setup-tools' ]))

from extended_setup import ExtendedSetupManager
ExtendedSetupManager('dataclasses_config').setup \
(
    name = 'dataclasses-config',
    url = 'https://gitlab.com/Hares-Lab/dataclasses-config',
    license = "BSD 2-Clause License",
    short_description = "A simple library extension for HOCON configs",
    min_python_version = '3.6.0',
    classifiers =
    [
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ]
)



from datetime import timedelta as TimeDelta, datetime as DateTime
from typing import *

from dateutil.relativedelta import relativedelta as RelativeDelta
from pytimeparse2 import parse as timeparse

from ..decorations import *


@deserialize_with(ConstructorDataType.Any)
class Duration(TimeDelta):
    """
    A helper class which represents any duration.
    This class is a subclass of standard timedelta.
    
    This class' constructor accepts either:
      
      - Single `int` or `float`: Counts this argument as the seconds
      - Single `str`: Uses the special library `pytimeparse` to parse the duration
      - Normal timedelta arguments (positional or keyword)
    
    Examples:
        ```python
        Duration(1000).total_seconds()              #   1000.0
        Duration('100ms').total_seconds()           #      0.1
        Duration('4 days 5 hours').total_seconds()  # 363600.0
        Duration('2:53').total_seconds()            #    173.0
        ```
    
    WARNING:
        
        1. This module PATCHES the `pytimeparse` library to support milliseconds,
        so it works correctly only with the version 1.1.18
        and only if no other module does so.
        This will be disabled as soon as the
        [Feature Request #22](https://github.com/wroberts/pytimeparse/issues/22) becomes closed.
        
        2. The `pytimeparse` library of version 1.1.8 DO NOT support
        large intervals, such as YEARS and MONTHS.
        This is the subject to to change.
        See [Feature Request #7](https://github.com/wroberts/pytimeparse/issues/7) for details.
    """
    
    @overload
    def __new__(cls: Type['Duration'], arg: str):
        pass
    @overload
    def __new__(cls: Type['Duration'], arg: TimeDelta):
        pass
    @overload
    def __new__(cls: Type['Duration'], arg: Union[int, float]):
        pass
    @overload
    def __new__(cls: Type['Duration'], *args, **kwargs):
        pass
    
    def __new__(cls: Type['Duration'], *args, **kwargs):
        if (len(args) == 1 and not kwargs):
            arg = args[0]
            
            if (isinstance(arg, RelativeDelta)):
                _now = DateTime.utcnow()
                arg = _now + arg - _now
            
            if (isinstance(arg, TimeDelta)):
                return super().__new__(cls, seconds=arg.total_seconds())
            elif (isinstance(arg, str)):
                return super().__new__(cls, seconds=timeparse(arg))
            elif (isinstance(arg, (int, float))):
                return super().__new__(cls, seconds=arg)
        
        return super().__new__(cls, *args, **kwargs)
    
    def __instancecheck__(self, instance):
        return isinstance(instance, TimeDelta)

__all__ = \
[
    'Duration',
]



"""
Helper package that contains classes for converting config data into program data (i.e., paths, durations, class refs, etc.)
"""

from .duration import *
from .dynamic_class import *
from .path import *

__all__ = [ ]
__pdoc__ = { }
__pdoc_extras__ = [ ]

submodules = \
[
    duration,
    dynamic_class,
    path,
]

for _m in submodules: __all__.extend(_m.__all__)
from .._helpers import create_documentation_index
create_documentation_index(submodules, __name__, __pdoc__, __pdoc_extras__)
del create_documentation_index, submodules
